#! /usr/bin/env python

########################################################################
# Lilith 1.0 version
#
# This file is the main file of Lilith
# made by J. Bernon and B. Dumont
#
# web page: http://lpsc.in2p3.fr/projects-th/lilith/
#
# in case of questions email bernon@lpsc.in2p3.fr dumont@lpsc.in2p3.fr
#
#
########################################################################

import sys, os.path, getopt, time, traceback
# import json # for storing experimental info---once this is implemented...

version = "1.0"
localdir = os.path.dirname(os.path.abspath(__file__))
default_exp_input_filepath = localdir + "/data/default.list"

try:
    from internal.ReadUserInput import *
    from internal.ReadExpInput import *
    from internal.ComputeLikelihood import *

    from internal.ComputeMuFromReducedCouplings import *
    from internal.ComputeReducedCouplings import *
    import internal.ReducedCouplingsLO
    import internal.ReducedCouplingsNNLO
    import internal.BR_SM

    import_check = 1
except:
    print "*******  Error while importing module for Lilith:  *******"
    exc_type, exc_value, exc_traceback = sys.exc_info()
    traceback.print_exception(exc_type, exc_value, exc_traceback, limit=2, file=sys.stdout)
    import_check = 0
    sys.exit()

#from internal.ComputeMuFromXsBr import *

class Lilith:
    """Main class. Read the experimental and user input and calculate the
    likelihood."""

    def __init__(self, verbose=False):
        """Initialization:
        initialize the relevant attributes, that's all."""

        if verbose:
            self.verbose = True
        else:
            self.verbose = False
        self.exp_input = None

        self.BR = {}
        self.eff_VVH = {}
        self.LO_FormFactors = {}
        self.redC_functions = {}
        self.grid_interpolation = {}


    def info(self, message):
        """print information only is verbose is True"""
        
        if self.verbose:
            print message

    def SetExpInput(self, exp_input_filepath=default_exp_input_filepath):
        # ------------------------
        # read experimental input
        # ------------------------
        filepath_split = exp_input_filepath.split("/")
        filename = filepath_split[-1]
        serialized_filepath = "/".join(filepath_split[:-1]) + "/_" + filename + ".json"

        # if the experimental input (list) file does not exist
        if not os.path.isfile(exp_input_filepath):
            sys.stderr.write("Error: the experimental input file \"" + \
                             exp_input_filepath + "\" does " + "not exist (or is a directory).\n")
            sys.exit()

        # if the serialized file does not exist
        # or if it has been modified after generation of serialized experimental input
        if not os.path.exists(serialized_filepath) or \
        os.path.getmtime(exp_input_filepath) > os.path.getmtime(serialized_filepath):
            self.info("Processing the experimental input...")
            
            if self.eff_VVH == {}:
                self.eff_VVH = internal.BR_SM.GetEffVVHfunctions()
            
            self.exp_input = ReadExpInput()

            # read (and check) each individual XML file from the experimental list file
            filelist = self.exp_input.get_filelist(exp_input_filepath)
            
            for expfile in filelist:
                self.exp_input.read_file(expfile, self.eff_VVH)

            """# then, check consistency of everything...
            self.exp_input.consistency_check()
            # ...and serialize the object in JSON format
            self.exp_input.serialize_json()
        else:
            # open the serialized experimental input
            print "Processing the experimental input..."
            with open(serialized_filepath, "r") as file:
                self.exp_input = json.load(file)"""

        # now, the experimental signal strengths are in self.exp_input.mu
        return True



    def GetLikelihood(self, user_input_filepath):
        try:
            if self.exp_input is None:
                sys.stderr.write("The experimental information is not found. Two possible reasons: \n"
                                 "1) Please check that your input file does exist (see #define Dataset in main.c) \n"
                                 "2) LilithInit() method should be called prior to "
                                 "GetLilithLikelihood().\n"
                                 "Lilith returns (-2LogL, ndf) = (-1, -1)\n")
                return (-1, -1, -1)
            #sys.exit()
            
            # ----------------
            # read user input
            # ----------------
            self.info("Reading the user input...")

            t0 = time.clock()
            user_input = ReadUserInput(user_input_filepath)
                        
            #print "reading the user input takes ", time.clock() - t0, "s"
            if not user_input.isInLightRange:
                  return (-1,-1,-1)
            

    ############################## Reduced couplings mode ##############################


            if user_input.mode == "reducedcouplings":
                self.info("Entering reduced couplings mode...\n")
                
                
                # if the SM BR grids have not been read yet
                if self.BR == {}:
                    self.BR = internal.BR_SM.GetBRfunctions()
                    for key, val in self.BR.items():
                        self.BR[key] = val(user_input.mh)
                
                
                # if the LO form factors functions have not been created yet
                if self.LO_FormFactors == {}:
                    if user_input.precision == "LO":
                        self.LO_FormFactors = internal.ReducedCouplingsLO.computeFormFactors()
                        for key, val in self.LO_FormFactors.items():
                            self.LO_FormFactors[key] = val(user_input.mh)
   
                # if the grids have not been read yet
                if self.grid_interpolation == {}:
                    if user_input.precision == "LO":
                        self.grid_interpolation.update(internal.ReducedCouplingsLO.GetVBF_LO())
                    else:
                        self.grid_interpolation.update(internal.ReducedCouplingsNNLO.Getgg_BESTQCD())
                        self.grid_interpolation.update(internal.ReducedCouplingsNNLO.GetGaGa_BESTQCD())
                        self.grid_interpolation.update(internal.ReducedCouplingsNNLO.GetZGa_BESTQCD())
                        self.grid_interpolation.update(internal.ReducedCouplingsNNLO.GetVBF_BESTQCD())
                        self.grid_interpolation.update(internal.ReducedCouplingsNNLO.GetggF_LHC_BESTQCD())
                        self.grid_interpolation.update(internal.ReducedCouplingsNNLO.GetggF_Tev_BESTQCD())

                    for key, val in self.grid_interpolation.items():
                        self.grid_interpolation[key] = val(user_input.mh)
                
            
                if self.redC_functions == {}:
                    
                    if user_input.precision == "LO":
                        self.redC_functions["redCggF_LO"] = internal.ReducedCouplingsLO.redCggF_LO
                    else:
                        self.redC_functions["redCggF_NNLO_LHC8"] = internal.ReducedCouplingsNNLO.redCggF_NNLO_LHC8
                        self.redC_functions["redCggF_NNLO_Tev"] = internal.ReducedCouplingsNNLO.redCggF_NNLO_Tev
                
                    if "gammagamma" not in user_input.redC:
                        if user_input.precision == "LO":
                            self.redC_functions["redCgammagamma_LO"] = internal.ReducedCouplingsLO.redCgammagamma_LO
                        else:
                            self.redC_functions["redCgammagamma_NNLO"] = internal.ReducedCouplingsNNLO.redCgammagamma_NNLO
            
                    if "gg" not in user_input.redC:
                        if user_input.precision == "LO":
                            self.redC_functions["redCgg_LO"] = internal.ReducedCouplingsLO.redCgg_LO
                        else:
                            self.redC_functions["redCgg_NNLO"] = internal.ReducedCouplingsNNLO.redCgg_NNLO
                                
                    if "Zgamma" not in user_input.redC:
                        if user_input.precision == "LO":
                            self.redC_functions["redCZgamma_LO"] = internal.ReducedCouplingsLO.redCZgamma_LO
                        else:
                            self.redC_functions["redCZgamma_NNLO"] = internal.ReducedCouplingsNNLO.redCZgamma_NNLO


                    if "VBF" not in user_input.redC:
                        if user_input.precision == "LO":
                            self.redC_functions["redCVBF_LO"] = internal.ReducedCouplingsLO.redCVBF_LO
                        else:
                            self.redC_functions["redCVBF_NLO"] = internal.ReducedCouplingsNNLO.redCVBF_NLO


                if "gg" not in user_input.redC or "gammagamma" not in user_input.redC or \
                        "Zgamma" not in user_input.redC or "VBF" not in user_input.redC or \
                        "ggF" not in user_input.redC:
                    # if reduced couplings need to be calculated for at least one of the
                    # loop-induced processes
                    
                    t0 = time.clock()
                    computed_couplings = ComputeReducedCouplings(user_input.redC,
                                                            user_input.precision,
                                                            user_input.mh,
                                                            self.LO_FormFactors,
                                                            self.redC_functions,
                                                            self.grid_interpolation)

                    #print "computing reduced couplings takes ", time.clock() - t0, "s"

                    self.info("The following reduced couplings have been computed at " + \
                                user_input.precision + " accuracy:")
                    for Cname, Cvalue in computed_couplings.new_redC.items():
                        self.info(Cname + " = " + str(Cvalue))
                    
                    # TODO: have an option to run the code just in order to have reduced
                    # couplings computation and nothing else
                    
                    
                    user_input.redC.update(computed_couplings.new_redC)
                    
                    
                    t0 = time.clock()
                    computed_mus = ComputeMuFromReducedCouplings(user_input.redC,
                                                                user_input.extraBR,
                                                                user_input.mh,
                                                                user_input.precision,
                                                                self.BR)
                    #print "computing mu from reduced couplings takes", time.clock() - t0, "s"
                    user_input.mu = computed_mus.mu

            elif user_input.mode == "signalstrengths":
                print "Entering signal strengths mode...\n"
                # nothing more to do, user_input.mu is already filled

            elif user_input.mode == "xsBR":
                print "Entering xs and BR mode...\n"

                computed_mus = ComputeMuFromXsBr(user_input.xs, user_input.BR,
                                                 user_input.precision, user_input.mh)
                user_input.mu = computed_mus.mu

            t0 = time.clock()
            likelihood = ComputeLikelihood()
            self.likelihoods = likelihood.likelihoods # likelihoods dict
            l = likelihood.compute(self.exp_input.mu, user_input.mu)
            #print "computing the likelihood takes ", time.clock() - t0, "s\n"
            return l
    
        except:
            if import_check == 1:
                print "********  Error in Lilith:  ********"
                exc_type, exc_value, exc_traceback = sys.exc_info()
                traceback.print_exception(exc_type, exc_value, exc_traceback, limit=2, file=sys.stdout)
                print "Lilith returns (-2LogL, ndf) = (-1, -1, -1)\n"

            return (-1, -1, -1)



    def GetCouplings(self, user_input_filepath):
        try:
            if self.exp_input is None:
                sys.stderr.write("The experimental information is not found. Two possible reasons: \n"
                                 "1) Please check that your input file does exist \n"
                                 "2) LilithInit() method should be called prior to "
                                 "GetCouplings().\n"
                                 "Lilith returns {} \n")
                return {}
            #sys.exit()
            
            # ----------------
            # read user input
            # ----------------
            self.info("Reading the user input...")

            t0 = time.clock()
            user_input = ReadUserInput(user_input_filepath)
                        
            #print "reading the user input takes ", time.clock() - t0, "s"
            if not user_input.isInLightRange:
              return {}
            

            if user_input.mode == "reducedcouplings":
                
                # if the SM BR grids have not been read yet
                if self.BR == {}:
                    self.BR = internal.BR_SM.GetBRfunctions()
                    for key, val in self.BR.items():
                        self.BR[key] = val(user_input.mh)
                
                
                # if the LO form factors functions have not been created yet
                if self.LO_FormFactors == {}:
                    if user_input.precision == "LO":
                        self.LO_FormFactors = internal.ReducedCouplingsLO.computeFormFactors()
                        for key, val in self.LO_FormFactors.items():
                            self.LO_FormFactors[key] = val(user_input.mh)
   
                # if the grids have not been read yet
                if self.grid_interpolation == {}:
                    if user_input.precision == "LO":
                        self.grid_interpolation.update(internal.ReducedCouplingsLO.GetVBF_LO())
                    else:
                        self.grid_interpolation.update(internal.ReducedCouplingsNNLO.Getgg_BESTQCD())
                        self.grid_interpolation.update(internal.ReducedCouplingsNNLO.GetGaGa_BESTQCD())
                        self.grid_interpolation.update(internal.ReducedCouplingsNNLO.GetZGa_BESTQCD())
                        self.grid_interpolation.update(internal.ReducedCouplingsNNLO.GetVBF_BESTQCD())
                        self.grid_interpolation.update(internal.ReducedCouplingsNNLO.GetggF_LHC_BESTQCD())
                        self.grid_interpolation.update(internal.ReducedCouplingsNNLO.GetggF_Tev_BESTQCD())

                    for key, val in self.grid_interpolation.items():
                        self.grid_interpolation[key] = val(user_input.mh)
                
            
                if self.redC_functions == {}:
                    
                    if user_input.precision == "LO":
                        self.redC_functions["redCggF_LO"] = internal.ReducedCouplingsLO.redCggF_LO
                    else:
                        self.redC_functions["redCggF_NNLO_LHC8"] = internal.ReducedCouplingsNNLO.redCggF_NNLO_LHC8
                        self.redC_functions["redCggF_NNLO_Tev"] = internal.ReducedCouplingsNNLO.redCggF_NNLO_Tev
                
                    if "gammagamma" not in user_input.redC:
                        if user_input.precision == "LO":
                            self.redC_functions["redCgammagamma_LO"] = internal.ReducedCouplingsLO.redCgammagamma_LO
                        else:
                            self.redC_functions["redCgammagamma_NNLO"] = internal.ReducedCouplingsNNLO.redCgammagamma_NNLO
            
                    if "gg" not in user_input.redC:
                        if user_input.precision == "LO":
                            self.redC_functions["redCgg_LO"] = internal.ReducedCouplingsLO.redCgg_LO
                        else:
                            self.redC_functions["redCgg_NNLO"] = internal.ReducedCouplingsNNLO.redCgg_NNLO
                                
                    if "Zgamma" not in user_input.redC:
                        if user_input.precision == "LO":
                            self.redC_functions["redCZgamma_LO"] = internal.ReducedCouplingsLO.redCZgamma_LO
                        else:
                            self.redC_functions["redCZgamma_NNLO"] = internal.ReducedCouplingsNNLO.redCZgamma_NNLO


                    if "VBF" not in user_input.redC:
                        if user_input.precision == "LO":
                            self.redC_functions["redCVBF_LO"] = internal.ReducedCouplingsLO.redCVBF_LO
                        else:
                            self.redC_functions["redCVBF_NLO"] = internal.ReducedCouplingsNNLO.redCVBF_NLO


                if "gg" not in user_input.redC or "gammagamma" not in user_input.redC or \
                        "Zgamma" not in user_input.redC or "VBF" not in user_input.redC or \
                        "ggF" not in user_input.redC:
                    # if reduced couplings need to be calculated for at least one of the
                    # loop-induced processes
                    
                    t0 = time.clock()
                    computed_couplings = ComputeReducedCouplings(user_input.redC,
                                                            user_input.precision,
                                                            user_input.mh,
                                                            self.LO_FormFactors,
                                                            self.redC_functions,
                                                            self.grid_interpolation)

                    #print "computing reduced couplings takes ", time.clock() - t0, "s"

                    self.info("The following reduced couplings have been computed at " + \
                                user_input.precision + " accuracy:")
                    for Cname, Cvalue in computed_couplings.new_redC.items():
                        self.info(Cname + " = " + str(Cvalue))
                    
                    user_input.redC.update(computed_couplings.new_redC)
                    
                return user_input.redC
                    
                  
            elif user_input.mode == "signalstrengths":
                print "GetCouplings() is only available in the reduced couplings mode"
                return {}
    
        except:
            if import_check == 1:
                print "********  Error in Lilith: lilith.GetCouplings()  ********"
                print "Please check that the user input file does exist"
                exc_type, exc_value, exc_traceback = sys.exc_info()
                traceback.print_exception(exc_type, exc_value, exc_traceback, limit=2, file=sys.stdout)

            return {}


################################################################################
################################################################################


if __name__ == "__main__": # if lilith.py is called directly
    def usage():
        print "Usage: ./lilith.py model_input_xml [experimental_input_list]"
        print "Example: ./lilith.py examples/input_mu.xml [data/default.list]"
    

    # at least one argument, check if asking for help
    if len(sys.argv) < 2 or (sys.argv[1] == "-h" or sys.argv[1] == "--help"):
        usage()
        sys.exit()

    user_input_filepath = sys.argv[1]

    # default values
    exp_input_filepath = default_exp_input_filepath
    verbose = False
    silent = False
    outputfilepath = ""

    # at least 2 args: set exp_input_filepath if not an option
    if len(sys.argv) >= 3:
        n = 2
        if sys.argv[2][0] != "-":
            n = 3
            exp_input_filepath = sys.argv[2]

        # check all possible options...
        try:
            opts, args = getopt.getopt(sys.argv[n:],
            "hvso:", ["help", "verbose", "silent", "output="])
        except getopt.GetoptError as err:
            print err
            usage()
            sys.exit()

        # and take into account these options
        for o, a in opts:
            if o in ("-v", "--verbose"):
                verbose = True
                if silent:
                    print o, "is incompatible with silent mode."
                    sys.exit()
            elif o in ("-s", "--silent"):
                silent = True
                if verbose:
                    print o, "is incompatible with verbose mode."
                    sys.exit()
            elif o in ("-h", "--help"):
                usage()
                sys.exit()
            elif o in ("-o", "--output"):
                outputfilepath = a
            else:
                print "unhandled option"
                sys.exit()

    if not silent:
        print "---------------------------"
        print "Lilith version " + version
        print "---------------------------"


    # turning to the calculation of the likelihood
    Lilithcalc = Lilith(verbose)

    with open(user_input_filepath) as f:
        user_input_string = f.read()

    Lilithcalc.SetExpInput(exp_input_filepath)

    like = Lilithcalc.GetLikelihood(user_input_string)

    if not silent:
        print "-2*log(L) =", round(like[0],6)
        print "ndf =", like[1]
        if outputfilepath != "":
            with open(outputfilepath, "w") as output:
                if like == (-1, -1, -1):
                   isInRange = 0
                else:
                   isInRange = 1
                output.write("BLOCK    Lilith\n")
                output.write("0        "+str(isInRange)+"           # 0: Point not in the [123,128] range, 1: Point in the range\n" )
                output.write("1        "+str(round(like[0],4))+"     # -2*LogL\n")
                output.write("2        "+str(like[1])+"          # ndf\n")
                output.close()
                print "written in file \"" + outputfilepath + "\"."
    else:
        if outputfilepath == "":
            pass
        else:
           with open(outputfilepath, "w") as output:
            if like == (-1, -1, -1):
             isInRange = 0
            else:
             isInRange = 1
            output.write("BLOCK    Lilith\n")
            output.write("0        "+str(isInRange)+"           # 0: Point not in the [123,128] range, 1: Point in the range\n" )
            output.write("1        "+str(round(like[0],4))+"     # -2*LogL\n")
            output.write("2        "+str(like[1])+"          # ndf\n")
            output.close()
            #print "written in file \"" + outputfilepath + "\"."
